
import netscape.application.Color;

/*
 * This represents a C "if-else" statement.
 */

public class DataIfElseStatement extends DataSelectionStatement
{
	/* Class variables */
  public final static String   IF                       = "if";
  public final static String   ELSE                     = "else";
  public final static String   OPEN_PAREN               = "(";
  public final static String   CLOSE_PAREN              = ")";
  public final static String   EXPRESSION_INDEX         = "ExpressionIndex";
  public final static String   IF_STATEMENT_INDEX       = "IfStatementIndex";
  public final static String   ELSE_STATEMENT_INDEX     = "ElseStatementIndex";
  public final static String   SEMICOLON                = ";";
  public final static String   EMPTY_COMPOUND_STATEMENT = "\n;";

  public       static Color    IF_ELSE_COLOR            = Color.orange;
  public final static int      IF_EXPRESSION_ONLY       = 4;


	/* Instance Variables */
  protected DataExpression          expression;
  protected DataStatement           elseStatement;
  protected int                     lineNumberOfElse;

	/* Instance Methods */
  public DataIfElseStatement()
  {
    expression       = null;
    elseStatement    = null;
    lineNumberOfElse = DataComponent.INVALID_LINE_NUMBER;
  }

  public DataExpression getExpression () { return expression; }

  public void           setExpression ( DataExpression theExpression )
  {
    if ( expression != null )
      expression . setParent ( null );

    expression = theExpression;

    if ( expression != null )
      expression . setParent ( this );
  }

  public DataCompoundStatement getIfStatement ()     { return getStatement(); }
  public void                  setIfStatement ( DataComponent theIfStatement )
  {
    setStatement ( theIfStatement );
  }

  public DataStatement getElseStatement ()     { return elseStatement; }
  public void          setElseStatement ( DataComponent theElseStatement )
  {
    if ( elseStatement != null )
      elseStatement . setParent ( null );

    if ( theElseStatement == null )
      elseStatement = null;

    else if (   ( theElseStatement instanceof DataCompoundStatement )
	     || ( theElseStatement instanceof DataIfElseStatement   ) )
      elseStatement = (DataStatement) theElseStatement;

    else
      elseStatement = new DataCompoundStatement ( theElseStatement );

    if ( elseStatement != null )
      elseStatement . setParent ( this );
  }


  public void setLineNumberOfElse ( int theLineNumberOfElse )
		   { lineNumberOfElse = theLineNumberOfElse; }

  public int  getLineNumberOfElse() { return lineNumberOfElse; }

  public boolean hasValidLineNumberOfElse()
  {
    return DataComponent.isValidLineNumber ( getLineNumberOfElse() );
  }



	/* Allow searching of our expression... */
  public boolean hasSubcomponentFraction ( String theString )
  {
    return super               . hasSubcomponentFraction ( theString )
      || (   ( getExpression() != null )
	  && ( getExpression() . hasSubcomponentFraction ( theString ) ) );
  }


  public int getChildStatementCount()
  {
    return super . getChildStatementCount()
      +  ( getElseStatement() != null ? 1 : 0 );
  }

  public DataStatement  getChildStatement ( int theIndex )
  {
    if ( theIndex == 0 )
    {
      if ( getIfStatement()        != null )
	return getIfStatement();
      else if ( getElseStatement() != null )
	return getElseStatement();
      else
	return super . getChildStatement ( theIndex );
    }

    if (   ( theIndex           == 1    )
	&& ( getIfStatement()   != null )
	&& ( getElseStatement() != null ) )
    {
      return getElseStatement();
    }

    return super . getChildStatement (
			  theIndex
			  - ( (   ( getIfStatement()   != null )
			       && ( getElseStatement() != null ) ) ? 1 : 0 ) );
  }



  public boolean addSecondaryChild(DataComponent theChildToAdd,
				   DataComponent theAddChildAfterThisComponent)
  {
	/* addSecondaryChild() only gets invoked from the GUI interface...
         * (So, if someone just dropped a compound-statement,
	 *  do the right thing...)
	 */
    if ( getElseStatement() == null )
    {
      if ( theChildToAdd instanceof DataIfElseStatement )
      {
	setElseStatement ( theChildToAdd );
	return true;
      }
      else
      {
	setElseStatement ( new DataCompoundStatement ( null ) );
      }
    }

    if ( getElseStatement() instanceof DataIfElseStatement )
    {
      setElseStatement ( new DataCompoundStatement ( getElseStatement() ) );
    }

    return ( (DataCompoundStatement) getElseStatement() )
               . addChild ( theChildToAdd, theAddChildAfterThisComponent );
  }


  public boolean removeChild ( DataComponent theChildToRemove )
  {
	/*  If it IS the Else statement... */
    if ( getElseStatement() == theChildToRemove )
    {
      setElseStatement ( null );
      return true;
    }

	/* If it is IN the Else statement */
    if (   ( getElseStatement() != null                          )
	&& ( getElseStatement() instanceof DataCompoundStatement )
	&& ( ( (DataCompoundStatement) getElseStatement() )
	       . getIndexOfSubcomponent ( theChildToRemove )
	     != DataComponent.INVALID_INDEX                    ) )

    {
      return getElseStatement() . removeChild ( theChildToRemove );
    }

	/* Otherwise... */
    return super . removeChild ( theChildToRemove );
  }


	/* Overriden from DataStatementWithSubstatements */
  public boolean isPartOfThisStatement ( DataComponent  theDataComponent )
  {
    return (   ( super . isPartOfThisStatement ( theDataComponent ) )
	    || (   ( theDataComponent   == getElseStatement() )
		&& ( getElseStatement() != null               )     ) );
  }




  public DataComponent newComponent ( )
  {
    DataIfElseStatement  ifElseStatement = new DataIfElseStatement();

    ifElseStatement . addSubcomponent ( "\n" );
    ifElseStatement . setIndex ( DataIfElseStatement.IF );
    ifElseStatement . addSubcomponent ( " " );
    ifElseStatement . setIndex ( DataIfElseStatement.OPEN_PAREN );
    ifElseStatement . setIndex ( DataIfElseStatement.CLOSE_PAREN );
    ifElseStatement . addSubcomponent ( "\n" );
    ifElseStatement . setIndex ( DataIfElseStatement.ELSE );

    return ifElseStatement;
  }


  public String getWarnString ( int      theObjectSubset,
				boolean  theIsGenerating )
  {
    return super . getWarnString ( theObjectSubset, theIsGenerating )
      + " or DataIfElseStatement.IF_EXPRESSION_ONLY ("
      + DataIfElseStatement.IF_EXPRESSION_ONLY + ")";
  }

  public boolean isValidObjectSubset ( int      theObjectSubset,
				       boolean  theIsGenerating )
  {
    if ( theObjectSubset == DataIfElseStatement.IF_EXPRESSION_ONLY )
      return true;
    else
      return super . isValidObjectSubset ( theObjectSubset, theIsGenerating );
  }


  public boolean isValid ( int theObjectSubsetToValidate )
  {
      /* We should check to make sure it parses later on. *DBA*/

    if ( theObjectSubsetToValidate == DataIfElseStatement.IF_EXPRESSION_ONLY )
      return getExpression() != null;

	/* Otherwise... */

	/* Simple checks that we can do here.. */
    if (   ( getExpression()  == null )
	|| ( getIfStatement() == null ) )
      return false;

    return super.isValid ( theObjectSubsetToValidate );
  }


  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );

    if ( theObjectSubsetToParse == DataIfElseStatement.IF_EXPRESSION_ONLY )
      return theTDLParser . parseExpressionObject ( 0 );
    else
      return theTDLParser . parseIfElse ( 0 );
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    DataComponent  aObject = theObjectToCopy;


    if (   ( theObjectSubsetToSet == DataIfElseStatement.IF_EXPRESSION_ONLY )
	&& ( theObjectToCopy instanceof DataExpression                      ) )
    {
      if (    ( theCanDestroyObject == false )
	  &&  ( aObject             != null  ) )
      {
	aObject =  aObject . copyComponent();
     
	if ( aObject == null )
	{
	  System.err.println ( "[DataIfElseStatement:setFrom]  Error:  "
			 + "Unable to replicate object:  " + theObjectToCopy );
	  return false;
	}
      }

      setExpression ( (DataExpression) aObject );
      return true;
    }


    if ( /*NOT*/ !  ( theObjectToCopy instanceof DataIfElseStatement ) )
    {
      System.err.println ( "[DataIfElseStatement:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataIfElseStatement  theIfElseStatementToCopy
			      = (DataIfElseStatement) theObjectToCopy;


    aObject = theIfElseStatementToCopy . expression;
    if (    ( theCanDestroyObject == false )
	&&  ( aObject             != null  ) )
    {
      aObject =  aObject . copyComponent();
     
      if ( aObject == null )
      {
	System.err.println ( "[DataIfElseStatement:setFrom]  ERROR:  "
			     + "Unable to replicate object:  "
			     + theIfElseStatementToCopy . expression );
	return false;
      }
    }
    setExpression ( (DataExpression) aObject );


    if ( theObjectSubsetToSet == DataIfElseStatement.IF_EXPRESSION_ONLY )
      return true;


    aObject = theIfElseStatementToCopy . elseStatement;
    if (    ( theCanDestroyObject == false )
	&&  ( aObject             != null  ) )
    {
      aObject =  aObject . copyComponent();
     
      if ( aObject == null )
      {
	System.err.println ( "[DataIfElseStatement:setFrom]  ERROR:  "
			     + "Unable to replicate object:  "
			     + theIfElseStatementToCopy . elseStatement );
	return false;
      }
    }
    setElseStatement ( aObject );


    return super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
			     theCanDestroyObject );
  }


  public void generateExpression ( DataDestination  theOutputDestination,
				   int              theObjectSubsetToGenerate,
				   boolean          theIsPartOfIfElseStatement)
  {
    boolean  needsLeadingSpace  = theIsPartOfIfElseStatement,
             needsTrailingSpace = theIsPartOfIfElseStatement;

    if ( getExpression() != null )
    {
      if ( getExpression() . hasLeadingWhitespaceSubcomponent() )
	needsLeadingSpace = false;
      if ( getExpression() . hasTrailingWhitespaceSubcomponent() )
	needsTrailingSpace = false;
    }

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex ( DataIfElseStatement.OPEN_PAREN );

    	/* Write any pre-EXPRESSION non-significant tokens */
    generateSubcomponents ( DataSwitchStatement.EXPRESSION_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate,
			    needsLeadingSpace );

	/* Write out expression, if we have one... */
    if ( getExpression() != null )
      getExpression() . generate ( theOutputDestination,
				   theObjectSubsetToGenerate );

	/* Write any pre-")" tokens */
    generateSubcomponents ( DataIfElseStatement.CLOSE_PAREN,
			    theOutputDestination,
			    theObjectSubsetToGenerate, 
			    needsTrailingSpace );
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
    int                    i;
    DataCompoundStatement  dataCompoundStatement;
    String                 spawnDestroyClauseString;

       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

		/* Handle subset case... */
    if ( theObjectSubsetToGenerate == DataIfElseStatement.IF_EXPRESSION_ONLY )
    {
      generateExpression ( theOutputDestination, DataComponent.ENTIRE_OBJECT,
			   false );
      return;
    }

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex();

	/* Generate any labels that we have... */
    generateLabels ( theOutputDestination, theObjectSubsetToGenerate,
		     DataIfElseStatement.IF );

	/* Write any pre-IF non-significant tokens */
    generateSubcomponents ( DataIfElseStatement.IF,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write our IF keyword */
    theOutputDestination . write ( DataIfElseStatement.IF );

	/* Write any pre-"(" non-significant tokens */
    generateSubcomponents ( DataIfElseStatement.OPEN_PAREN,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write "(" */
    theOutputDestination . write ( DataIfElseStatement.OPEN_PAREN );


	/* Write the pre-expression tokens, the expression, */
	/* and the post-expression-pre-close-paren tokens   */
    generateExpression ( theOutputDestination, theObjectSubsetToGenerate,
			 true );


	/* Write ")" */
    theOutputDestination . write ( DataIfElseStatement.CLOSE_PAREN );


	/* Write any pre--IF-STATEMENT non-significant tokens */
    generateSubcomponents ( DataIfElseStatement.IF_STATEMENT_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );


	/* Try to recover gracefully if we have a problem... */
    if ( getIfStatement() == null )
    {
      setIfStatement ( new DataCompoundStatement ( null ) );
    }

	/* Add in the else-spawn-destroy clause */
    if (   ( isCxxSubset ( theObjectSubsetToGenerate ) )
	&& ( getElseStatement() != null                ) )
    {
      spawnDestroyClauseString
	= getElseStatement() . getSpawnDestroyClause ( false );

      if ( DataComponent.isEmptyString ( spawnDestroyClauseString ) == false )
      {
	getIfStatement() . addSubcomponentBeforeCloseBrace (
	   new DataComponentPlaceholder (
		 "\n"
		 + DataComponent.getSpaceString (
				    theOutputDestination.getSecondaryIndent() )
		 + spawnDestroyClauseString ) );
      }
    }

	/* Write the IF-STATEMENT */
    getIfStatement() . generate ( theOutputDestination,
				  theObjectSubsetToGenerate );

	/* Remove the else-spawn-destroy clause */
    if (   ( isCxxSubset ( theObjectSubsetToGenerate ) )
	&& ( getElseStatement() != null                ) )
    {
      getIfStatement() . removeSubcomponentBeforeCloseBrace();
    }



	/* Do we have an else clause? */
    if (   ( getElseStatement() != null )
	|| ( isCxxSubset ( theObjectSubsetToGenerate ) ) )
    {
	/* Write the line number of the else keyword */
      generateRelevantLineNumberMacros ( theOutputDestination,
					 theObjectSubsetToGenerate,
					 getLineNumberOfElse(),
					 DataIfElseStatement.ELSE );

	/* Write any pre-ELSE non-significant tokens */
      generateSubcomponents ( DataIfElseStatement.ELSE,
			      theOutputDestination,
			      theObjectSubsetToGenerate, false );

	/* Write our ELSE keyword */
      theOutputDestination . write ( DataIfElseStatement.ELSE );

	/* Write any pre--ELSE-STATEMENT non-significant tokens */
      generateSubcomponents ( DataIfElseStatement.ELSE_STATEMENT_INDEX,
			      theOutputDestination,
			      theObjectSubsetToGenerate, false );


	/* Handle Cxx case. */
      if ( isCxxSubset ( theObjectSubsetToGenerate ) )
      {
	    /* Find us a compound-statement for the else clause */
	if ( getElseStatement() != null )
	{
	  if ( getElseStatement() instanceof DataCompoundStatement )
	    dataCompoundStatement = (DataCompoundStatement) getElseStatement();
	  else
	  {
	    dataCompoundStatement
	      = new DataCompoundStatement ( getElseStatement() );
	    dataCompoundStatement . setParent ( this );
	  }
	}
	else /* getElseStatement() == null */
	{
	  dataCompoundStatement = new DataCompoundStatement ( null );
	  dataCompoundStatement . setParent ( this );
	}
	
	  /* Add in our if-spawn-destroy clause */
	spawnDestroyClauseString
	  = getIfStatement() . getSpawnDestroyClause ( false );

	if ( DataComponent.isEmptyString( spawnDestroyClauseString ) == false )
	{
	  dataCompoundStatement . addSubcomponentAfterOpenBrace (
	     new DataComponentPlaceholder (
		   "\n" 
		 + DataComponent.getSpaceString (
				    theOutputDestination.getSecondaryIndent() )
		 + spawnDestroyClauseString ) );
	}


	dataCompoundStatement . generate ( theOutputDestination,
					   theObjectSubsetToGenerate );

	  /* Remove the else-spawn-destroy clause */
	dataCompoundStatement . removeSubcomponentAfterOpenBrace();

      } /* IF ( isCxxSubset ( theObjectSubsetToGenerate ) ) */

      else /* getElseStatement() != null && NOT a CxxSubset */
      {
	  /* Write our ELSE-STATEMENT */
	getElseStatement() . generate ( theOutputDestination,
				      theObjectSubsetToGenerate );

      }

    } /* Generating ELSE clause */


	/* Write any remaining non-significant tokens */
	/* (There *SHOULD* *NOT* be any...  But just in case...) */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );
  }



	/********************************************************/
        /********** Methods to interface with Graphing **********/
	/********************************************************/
  public boolean supportsGraphType ( int theGraphType )
  {
    if (   ( theGraphType == GraphRoot.GUI_TREE_GRAPH ) )
      return true;
    else
      return false;
  }

  public GraphModule createStatementGraphModule ( int theGraphType )
  {
    if (   ( theGraphType != GraphRoot.GUI_TREE_GRAPH ) )
      return null;

    /* else ... */

    /* Building:  +- if - [.]
     *            |   |
     *            |   +--then
     *            |   |    |
     *            |   |    
     *            |   +--else
     *            |        |
     */

    GraphConnection elseTail = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );

    GraphNode       elseNode = new GraphNode ( new SimpleLabel ( "ELSE" ),
				             GraphComponent.LEFT,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, elseTail );

    GraphConnection toElse = new GraphConnection ( GraphComponent.LEFT,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, false, elseNode );

    GraphMeet       elseMeet = new GraphMeet ( GraphComponent.UP,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     GraphComponent.LEFT, toElse );

    GraphConnection toElseMeet = new GraphConnection ( GraphComponent.UP,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, true, elseMeet );


    GraphConnection thenTail = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );

    GraphNode       thenNode = new GraphNode ( new SimpleLabel ( "THEN" ),
				             GraphComponent.LEFT,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, thenTail );

    GraphConnection toThen = new GraphConnection ( GraphComponent.LEFT,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, false, thenNode );

    GraphMeet       thenMeet = new GraphMeet ( GraphComponent.UP,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     GraphComponent.LEFT, toThen );

    thenMeet . addChild ( toElseMeet );


    GraphConnection toThenMeet = new GraphConnection ( GraphComponent.UP,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, true, thenMeet );



    ScrollTextOfDataComponent  ifExpressionText
      = new ScrollTextOfDataComponent( this,
				       DataIfElseStatement.IF_EXPRESSION_ONLY);

    GraphNode       expressionNode = new GraphNode ( ifExpressionText,
					     GraphComponent.LEFT,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     true, null );

    GraphConnection toExpression = new GraphConnection ( GraphComponent.LEFT,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, false, expressionNode);

    GraphNode       ifNode = new GraphNode ( new SimpleLabel ( "IF" ),
				             GraphComponent.LEFT,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, toExpression );

    ifNode . addChild ( toThenMeet );


    GraphConnection toIfNode = new GraphConnection ( GraphComponent.LEFT,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     false, false, ifNode );
    

    GraphMeet       startMeet = new GraphMeet ( GraphComponent.UP,
					     DataIfElseStatement.IF_ELSE_COLOR,
					     GraphComponent.LEFT,
					     toIfNode );

    GraphConnection tail = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );
    startMeet . addChild ( tail );

    GraphModule  graphModule = new GraphModule ( startMeet, tail,
						 thenTail, elseTail,
						 this, true, theGraphType );

    updateGraphModule ( this, null, graphModule );

    return graphModule;
  }



	/* Invoked by this.update(*) invoking GraphModule.update() */
  public boolean updateGraphModule ( DataComponent  theModifiedComponent,
				     Object         theStoredData,
				     GraphModule    theGraphModule )
  {
    boolean  returnValue = true;

	/* Skip updates to parents... */
    if ( isPartOfThisStatement ( theModifiedComponent ) == false )
      return true;

	/* Deal with If statement */
    if (   ( getIfStatement() != null )
	&& ( getIfStatement() . updateGraphModule( theGraphModule,
						   GraphModule.CHILD_COMPONENT)
	     == false )
        )
    {
      returnValue = false;
    }


	/* Deal with Else when it's a compound-statement */
    if (   ( getElseStatement() != null )
	&& ( getElseStatement() instanceof DataCompoundStatement )
	&& ( ((DataCompoundStatement) getElseStatement())
	       . updateGraphModule ( theGraphModule,
				     GraphModule.SECONDARY_CHILD_COMPONENT )
	     == false ) )
    {
      returnValue = false;
    }

	/* If ELSE is an if-else statement, find its GraphModule & recurse. */
    else if (   ( getElseStatement() == null )
	     || ( getElseStatement() instanceof DataIfElseStatement ) )
    {
      GraphModule        graphModule;
      GraphConnection    graphConnection = null;
      GraphComponent     graphComponent  = null;

      if (   ( theGraphModule . getSecondaryChildComponent() 
	       instanceof GraphConnection )
	   == false )
      {
	System.err.println (
	    "[DataIfElseStatement:updateGraphModule]  Error:  "
	    + "theGraphModule.getSecondaryChildComponent() is not a "
	    + "GraphConnection.   Aborting..." );
	return false;
      }

	/* Grab that connection */
      graphConnection
	= (GraphConnection) (theGraphModule . getSecondaryChildComponent());

	/* Find it's destination -- the first element of the next graphModule*/
      graphComponent = graphConnection . getDestination();


	/* If we have a next component */
      if ( graphComponent != null )
      {
	   /* Find it's graphModule */
	graphModule = graphComponent . getOwningGraphModule();

	if (   ( graphModule != null           )
	    && ( graphModule != theGraphModule ) )
	{
		/* Is it a wrong DataComponent? */
	  if (   ( getElseStatement()               == null               )
	      || ( graphModule . getDataComponent() != getElseStatement() ) )
	  {
	    graphModule . removeModuleFromGraph();
	    graphComponent = null;
	  }

		/* If it's the right thing, just recurse down... */
	  else if (   ( getElseStatement() != null )
		   && ( ( (DataIfElseStatement) getElseStatement() )
			  . updateGraphModule ( getElseStatement(),
						null, graphModule  )
			== false ) )
	  {
	    returnValue = false;
	  }
	}
	else /* ( graphModule == null || graphModule = theGraphModule ) */
	{
	  System.err.println (
	      "[DataIfElseStatement:updateGraphModule]  Error:  "
	      + "Consistency error in Else Graph Components.  (\""
	      + graphModule + "\").   Attempting to recover..." );

	    /* Kill the current graph-connection */
	  graphConnection . setDestination ( null );

	    /* Do what we normally do without a graphConnection destination. */
	  graphComponent = null;
	}
      }

	/* If we didn't have a graphComponent, or we just did a delete... */
      if (   ( graphComponent     == null )
	  && ( getElseStatement() != null ) )
      {
	  /* Create a module for this DataComponent */
	graphModule = getElseStatement() 
		       . createGraphModule( theGraphModule . getGraphType() );

	  /* Insert this new module */
	graphModule . insertModuleAt ( graphConnection );
      }
    }

    return returnValue;
  }

}
