
import netscape.application.Color;

/*
 * This represents a break, continue, or return C++ statement,
 * and the SUCCESS, FAIL, BYPASS, and POSTPONE TDL statements.
 */

public class DataJumpStatement extends DataStatement 
{
	/* Class variables */
  public final static String   FIRST_TOKEN_INDEX = "First";
  public final static String   SEMICOLON         = ";";

  public final static int      BREAK_TYPE        = 0;
  public final static int      CONTINUE_TYPE     = 1;
  public final static int      RETURN_TYPE       = 2;
  public final static int      SUCCESS_TYPE      = 3;
  public final static int      FAIL_TYPE         = 4;
  public final static int      BYPASS_TYPE       = 5;
  public final static int      POSTPONE_TYPE     = 6;

  public final static String   JUMP_TYPES[]      = { "break",
						     "continue",
						     "return",
						     "SUCCESS",
						     "FAIL",
						     "BYPASS",
						     "POSTPONE" };


  public       static Color    JUMP_COLOR        = Color.pink;

  public final static int      PARSE_GUI_JUMP_STATEMENT = 5003;


	/* Instance Variables */
  protected int              type;
  protected DataExpression   dataExpression; /* Return value. Should be null.*/

	/* FAIL's have an associated exception.  The DataSpawnTask class  */
	/* has all the features we need to support this exception (here). */
  protected DataSpawnTask          exceptionTask;


	/* Instance Methods */
  public DataJumpStatement()
  {
    type           = DataJumpStatement.RETURN_TYPE;
    dataExpression = null;
    exceptionTask  = null;
  }

  public int  getType() { return type; }

  public void setType ( int theType )
  {
    if (   ( theType == DataJumpStatement.BREAK_TYPE    )
	|| ( theType == DataJumpStatement.CONTINUE_TYPE )
	|| ( theType == DataJumpStatement.RETURN_TYPE   )
	|| ( theType == DataJumpStatement.SUCCESS_TYPE  )
	|| ( theType == DataJumpStatement.FAIL_TYPE     )
	|| ( theType == DataJumpStatement.BYPASS_TYPE   )
	|| ( theType == DataJumpStatement.POSTPONE_TYPE ) )
    {
      type = theType;
    }
    else
    {
      System.err.println( "[DataJumpStatement:setType]  Error:  "
			  + "Illegal type: " + theType + "  Aborting set..." );
    }
  }

  public DataExpression getDataExpression () { return dataExpression; }

  public void           setDataExpression ( DataExpression theDataExpression )
  {
    if ( dataExpression != null )
      dataExpression . setParent ( null );

    dataExpression = theDataExpression;

    if ( dataExpression != null )
      dataExpression . setParent ( this );
  }


  public DataSpawnTask getExceptionTask () { return exceptionTask; }
  public boolean       setExceptionTask ( DataSpawnTask theExceptionTask )
  {
    if (   (   ( theExceptionTask                            != null )
	    && ( theExceptionTask . getTaskName()            != null )
	    && ( theExceptionTask . getTaskName() . length() >  0    )
	    && ( theExceptionTask . getConstraintCount()     == 0    ) )
	|| (     theExceptionTask == null                              ) )
    {
      exceptionTask = theExceptionTask;
      return true;
    }
    else
    {
      System.err.println ( "[DataJumpStatement:setExceptionTask]  Error: "
			   + "Bad object for theExceptionTask:  " 
			   + theExceptionTask . toString() );
      return false;
    }
  }



	/* Allow searching of our expression... */
  public boolean hasSubcomponentFraction ( String theString )
  {
    return super                   . hasSubcomponentFraction ( theString )
      || (   ( getDataExpression() != null )
	  && ( getDataExpression() . hasSubcomponentFraction ( theString ) ) );
  }


  public DataComponent newComponent ( )
  {
    DataJumpStatement  statement = new DataJumpStatement();

    statement . addSubcomponent ( "\n" );
    statement . setIndex ( DataJumpStatement.FIRST_TOKEN_INDEX );
    statement . setIndex ( DataJumpStatement.SEMICOLON         );

    return statement;
  }



	/* Validates code that is inside of a Task */
  public void validateInternalCode( int                         theReference,
				    DataValidateCodeReturnValue theReturnValue)
  {
    DataComponent        ancestor;
    DataTaskDefinition   parentTask = getParentTaskDefinition();

    super . validateInternalCode ( theReference, theReturnValue );


    if ( getType() == DataJumpStatement.BREAK_TYPE )
    {
      for ( ancestor = getParent();
	    (   (   ancestor != null )
	     && ( ( ancestor instanceof DataIterationStatement ) == false )
	     && ( ( ancestor instanceof DataSwitchStatement    ) == false ) );
	    ancestor = ancestor . getParent() )
        ; /* NULL (EMPTY) LOOP BODY */

      if ( ancestor == null )
      {
	theReturnValue
	  . addWarning ( getLineNumberString() )
	  . write ( "Break statement outside of any enclosing iteration "
		    + "or switch statement.\n" );
      }
    }

    if ( getType() == DataJumpStatement.CONTINUE_TYPE )
    {
      for ( ancestor = getParent();
	    (   (   ancestor != null )
	     && ( ( ancestor instanceof DataIterationStatement ) == false ) );
	    ancestor = ancestor . getParent() )
        ; /* NULL (EMPTY) LOOP BODY */

      if ( ancestor == null )
      {
	theReturnValue
	  . addWarning ( getLineNumberString() )
	  . write ( "Continue statement outside of any enclosing iteration "
		    + "statement.\n" );
      }
    }


    if (   ( getType()          == DataJumpStatement.FAIL_TYPE           )
	&& ( getExceptionTask() != null                                  )
	&& (   ( getExceptionTask() . getTaskName()            == null )
	    || ( getExceptionTask() . getTaskName() . length() <= 0    ) ) )
    {
      theReturnValue
	. addError ( getLineNumberString() )
	. write ( "Exception specified in FAIL statement has no name.\n" );
    }

    if ( parentTask == null )
    {
      theReturnValue
	. addWarning ( getLineNumberString() )
	. write ( "Internal Error -- Unable to locate parent Task in ")
	. write ( "DataJumpStatement.validateInternalCode().\n" );
    }
    else
    {
      if (   ( getType()                  == DataJumpStatement.FAIL_TYPE    )
	  && ( getExceptionTask()         == null                           )
	  && ( parentTask . getTaskType() != DataTaskDefinition.HANDLER_TASK) )
      {
	theReturnValue
	  . addWarning ( getLineNumberString() )
	  . write ( "FAIL statement lacks an Exception.\n" );
      }

      if (   ( getType()                  == DataJumpStatement.BYPASS_TYPE  )
	  && ( parentTask . getTaskType() != DataTaskDefinition.HANDLER_TASK) )
      {
	theReturnValue
	  . addError ( getLineNumberString() )
	  . write ( "BYPASS statement encountered outside of an " )
	  . write ( "Exception-Handler Task.\n" );
      }
    }
  }



  public String getWarnString ( int      theObjectSubset,
				boolean  theIsGenerating )
  {
    return super . getWarnString ( theObjectSubset, theIsGenerating )
	      + " or DataJumpStatement.PARSE_GUI_JUMP_STATEMENT ("
	      + DataJumpStatement.PARSE_GUI_JUMP_STATEMENT + ")";
  }

  public boolean isValidObjectSubset ( int      theObjectSubset,
				       boolean  theIsGenerating )
  {
    if ( theObjectSubset == DataJumpStatement.PARSE_GUI_JUMP_STATEMENT )
      return true;
    else
      return super . isValidObjectSubset ( theObjectSubset, theIsGenerating );
  }


  public boolean isValid ( int theObjectSubsetToValidate )
  {
    DataTaskDefinition   parentTask = getParentTaskDefinition();

    if (   ( getType()          == DataJumpStatement.FAIL_TYPE           )
	&& ( getExceptionTask() != null                                  )
	&& (   ( getExceptionTask() . getTaskName()            == null )
	    || ( getExceptionTask() . getTaskName() . length() <= 0    ) ) )
    {
      return false;
    }

    if (   ( getType()          == DataJumpStatement.FAIL_TYPE             )
	&& ( getExceptionTask() == null                                    )
	&& ( parentTask         != null                                    )
	&& ( parentTask . getTaskType() != DataTaskDefinition.HANDLER_TASK ) )
    {
      return false;
    }

    return super . isValid ( theObjectSubsetToValidate );
  }


  protected
    DataComponent parseSpecificComponent ( TDLParser theTDLParser,
					   int       theObjectSubsetToParse )
    throws Exception
  {
    warnIfInvalidObjectSubset ( theObjectSubsetToParse,
				"parseSpecificComponent", false );
    return theTDLParser . parseJumpStatement ( 0 );
  }


  public boolean setFrom ( DataComponent  theObjectToCopy,
			   int            theObjectSubsetToSet,
			   boolean        theCanDestroyObject  )
  {
    if ( /*NOT*/ !  ( theObjectToCopy instanceof DataJumpStatement ) )
    {
      System.err.println ( "[DataJumpStatement:setFrom]  Error:  "
			   + "Illegal object.  (\""
			   + theObjectToCopy . getClass() . getName()
			   + "\")" );
      return false;
    }

    DataJumpStatement  theJumpStatementToCopy
			      = (DataJumpStatement) theObjectToCopy;
    DataComponent      aObject
			      = theJumpStatementToCopy.dataExpression;
    boolean            returnValue;


    if (   ( theCanDestroyObject == false )
	&& ( aObject             != null  ) )
    {
      aObject =  aObject . copyComponent();
     
      if ( aObject == null )
      {
	System.err.println ( "[DataJumpStatement:setFrom]  ERROR:  "
			     + "Unable to replicate object:  "
			     +  theJumpStatementToCopy.dataExpression );
	return false;
      }
    }

    setDataExpression ( (DataExpression) aObject );

    type = theJumpStatementToCopy . type;

    if (   theCanDestroyObject
	|| ( theJumpStatementToCopy.exceptionTask == null ) )
    {
      exceptionTask = theJumpStatementToCopy.exceptionTask;
    }
    else
    {
      exceptionTask
	= ( (DataSpawnTask) 
	    (theJumpStatementToCopy.exceptionTask . copyComponent()) );
    }

    returnValue =  super . setFrom ( theObjectToCopy, theObjectSubsetToSet,
				     theCanDestroyObject );

	/* Kludge to add "\n"'s to GUI-parsed jump-statements */
    if (   ( returnValue                                      == true )
	&& ( theObjectSubsetToSet
	        == DataJumpStatement.PARSE_GUI_JUMP_STATEMENT         )
	&& ( getIndex ( DataJumpStatement.FIRST_TOKEN_INDEX ) == 0    ) )
    {
      addSubcomponent ( "\n", 0 );

      setIndex (            DataJumpStatement.FIRST_TOKEN_INDEX,
		 getIndex ( DataJumpStatement.FIRST_TOKEN_INDEX ) + 1 );

      setIndex (            DataJumpStatement.SEMICOLON,
		 getIndex ( DataJumpStatement.SEMICOLON ) + 1 );
    }

    return returnValue;
  }


  public void generateReturn ( DataDestination  theOutputDestination,
			       int              theObjectSubsetToGenerate,
			       boolean          theDoCompletedSuccessfully,
			       boolean          theReturnIsPartOfTdlCode    )
  {
    if ( theDoCompletedSuccessfully )
    {
	/* Deal with #line macros */
      theOutputDestination . setUsingTdlFileName ( false );
      theOutputDestination . write ( "\n" );

      theOutputDestination
	. write ( DataComponent.CXX_TCM_TASK_COMPLETED_SUCCESSFULLY );
      theOutputDestination . write ( " ( " );
      theOutputDestination . write ( DataComponent.CXX_ENCLOSING_TASK_REF );
      theOutputDestination . write ( " );" );
    }

	/* Deal with #line macros */
    if ( theReturnIsPartOfTdlCode )
    {
      theOutputDestination . setUsingTdlFileName ( true );
      theOutputDestination . makeNextLineNumber ( getLineNumber() );
    }
    else
    {
      theOutputDestination . setUsingTdlFileName ( false );
    }
    theOutputDestination . write ( "\n" );


    theOutputDestination . write ( "return" );

    if ( getDataExpression() != null )
      getDataExpression() . generate ( theOutputDestination,
				       theObjectSubsetToGenerate );
  }


  public void generate ( DataDestination  theOutputDestination,
			 int              theObjectSubsetToGenerate )
  {
    int                  indent, unindent;
    DataTaskDefinition   parentTask = getParentTaskDefinition();

	/* Hack to deal with GUI-parsing... */
    if ( theObjectSubsetToGenerate
	   == DataJumpStatement.PARSE_GUI_JUMP_STATEMENT )
      theObjectSubsetToGenerate = DataComponent . ENTIRE_OBJECT;

       	/* Print a warning if necessary... */
    warnIfInvalidObjectSubset ( theObjectSubsetToGenerate, "generate", true );

    	/* Initialize us to generate non-significant tokens... */
    initializeGenerateSubcomponentIndex();

	/* Generate any labels that we have... */
    generateLabels ( theOutputDestination, theObjectSubsetToGenerate,
		     DataJumpStatement.FIRST_TOKEN_INDEX );

	/* Write any pre-first-token non-significant tokens */
    generateSubcomponents ( DataJumpStatement.FIRST_TOKEN_INDEX,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );


	/* Write the jump-statement's token(s) */

    if (   ( getType() == DataJumpStatement.BREAK_TYPE    )
	|| ( getType() == DataJumpStatement.CONTINUE_TYPE ) )
    {
      theOutputDestination
	. write ( DataJumpStatement.JUMP_TYPES [ getType() ] );
    }


    else if ( getType() == DataJumpStatement.RETURN_TYPE )
    {
      generateReturn ( theOutputDestination, theObjectSubsetToGenerate,
		       isCxxSubset ( theObjectSubsetToGenerate ), true );
    }


    else if (   (     getType()  == DataJumpStatement.SUCCESS_TYPE        )
	     || (     getType()  == DataJumpStatement.BYPASS_TYPE         )
	     || (     getType()  == DataJumpStatement.POSTPONE_TYPE       )

		/* FAIL implements as BYPASS under special circumstances */
	     || (   ( getType()          == DataJumpStatement.FAIL_TYPE )
		 && ( getExceptionTask() == null                        )
		 && ( parentTask         != null                        )
		 && ( isCxxSubset ( theObjectSubsetToGenerate )         )
		 && (    parentTask . getTaskType()
		      == DataTaskDefinition.HANDLER_TASK                ) ) )
    {
      if ( isCxxSubset ( theObjectSubsetToGenerate ) )
      {
	theOutputDestination . write ( "// " );
      }

      theOutputDestination
	. write ( DataJumpStatement.JUMP_TYPES [ getType() ] );

      if ( isCxxSubset ( theObjectSubsetToGenerate ) )
      {
	theOutputDestination . write ( " ;\n" );

	if (   ( getType() == DataJumpStatement.BYPASS_TYPE )
	    || ( getType() == DataJumpStatement.FAIL_TYPE   ) )
	{
	  theOutputDestination . write ( DataComponent.CXX_TCM_BYPASS );
	  theOutputDestination . write ( " ( " );
	  theOutputDestination . write ( DataComponent.CXX_ENCLOSING_TASK_REF);
	  theOutputDestination . write ( " );\n" );
	}

	generateReturn ( theOutputDestination, theObjectSubsetToGenerate,
			 getType() == DataJumpStatement.SUCCESS_TYPE, false );
      }
    }


    else if ( getType() == DataJumpStatement.FAIL_TYPE )
    {
      if ( isCxxSubset ( theObjectSubsetToGenerate ) )
      {
	theOutputDestination . setNewlineText ( "// " );
	theOutputDestination . write ( "// " );
      }

      theOutputDestination
	. write ( DataJumpStatement.JUMP_TYPES [ getType() ] );


      if ( getExceptionTask() != null )
      {
	   /* Write the exception-task, properly indenting it... */
	indent = theOutputDestination.indentToCurrentColumn();

	  /* Remove indentation for "// " string */
	if ( indent >= theOutputDestination . getNewlineText() . length() )
	{
	  theOutputDestination
	    . removeIndent ( theOutputDestination . getNewlineText()
			                          . length() );
	  indent -= theOutputDestination . getNewlineText() . length();
	}

	getExceptionTask() . generate ( theOutputDestination,
					DataSpawnTask.TASK_ONLY );
	theOutputDestination . removeIndent ( indent );
      }


      if ( isCxxSubset ( theObjectSubsetToGenerate ) )
      {
	theOutputDestination . write ( " ;" );
	theOutputDestination . clearNewlineText();

	/* Deal with #line macros */
	theOutputDestination . setUsingTdlFileName ( false );
	theOutputDestination . write ( "\n" );

	theOutputDestination . write ( DataComponent.CXX_TCM_FAIL );
	theOutputDestination . write ( " ( " );

	indent = theOutputDestination.indentToCurrentColumn();

	theOutputDestination . write ( DataComponent.CXX_ENCLOSING_TASK_REF);
	theOutputDestination . write ( ",\n" );


		/* Write the string-matcher bit */
	if ( getExceptionTask() != null )
	{
	  theOutputDestination
	    . write ( DataComponent.CXX_EXCEPTION_NAME_LEAD );
	  theOutputDestination . write ( getExceptionTask() . getTaskName() );
	}
	else  /* Generic exception case... */
	{
	  theOutputDestination
	    . write ( DataComponent.CXX_EXCEPTION_BASE_CLASS );
	}
	theOutputDestination . write ( "::" );
	theOutputDestination
	  . write ( DataComponent.CXX_EXCEPTION_MATCHES_METHOD_NAME );
	theOutputDestination . write ( ",\n" );


		/* Write the create-exception bit. */
	theOutputDestination
	  . write ( DataComponent.CXX_CREATE_EXCEPTION_FUNCTION_LEAD );

	theOutputDestination . addIndent (
	          DataComponent.CXX_CREATE_EXCEPTION_FUNCTION_LEAD . length());

	indent += DataComponent.CXX_CREATE_EXCEPTION_FUNCTION_LEAD . length();

	theOutputDestination . setStripLeadingWhitespace();

	if ( getExceptionTask() != null )
	{
		/* Unindent by the exception-tasks's leading whitespace */
	  unindent
	    = getExceptionTask()
	        . getLeadingWhitespaceIndent ( DataSpawnTask.TASK_NAME_INDEX );
	  if ( theOutputDestination . getIndent() > unindent )
	  {
	    theOutputDestination . removeIndent ( unindent );
	    indent -= unindent;
	  }

	  getExceptionTask()
	    . generateTask ( theOutputDestination,
			     DataComponent.ENTIRE_OBJECT,
			     false,  /* No leading space    */
			     false,  /* No leading comments */
			     true,   /* Indent nicely,      */
			     false,  /*   on current line.  */
			     isCxxSubset ( theObjectSubsetToGenerate ) );
	}
	else  /* Generic exception case... */
	{
	  theOutputDestination
	    . write ( DataComponent.CXX_EXCEPTION_BASE_CLASS );
	  theOutputDestination . write ( "()" );
	}

	theOutputDestination . removeIndent ( indent );

	theOutputDestination . write ( " );\n" );
	generateReturn ( theOutputDestination, theObjectSubsetToGenerate,
			 false, false );
      }
    }


    else
    {
      System.err.println( "[DataJumpStatement:generate]  Error:  "
			  + "Illegal type: " + getType() );
    }


	/* Write any non-significant tokens before the semicolon... */
    generateSubcomponents ( DataJumpStatement.SEMICOLON,
			    theOutputDestination,
			    theObjectSubsetToGenerate, false );

	/* Write our semicolon */
    theOutputDestination . write ( DataJumpStatement.SEMICOLON );

	/* Write any remaining non-significant tokens */
	/* (There *SHOULD* *NOT* be any...  But just in case...) */
    generateAllRemainingSubcomponents ( theOutputDestination,
					theObjectSubsetToGenerate, false );
  }



	/********************************************************/
        /********** Methods to interface with Graphing **********/
	/********************************************************/
  public boolean supportsGraphType ( int theGraphType )
  {
    if (   ( theGraphType == GraphRoot.GUI_TREE_GRAPH ) )
      return true;
    else
      return false;
  }

  public GraphModule createGraphModule ( int theGraphType )
  {
    if (   ( theGraphType != GraphRoot.GUI_TREE_GRAPH ) )
      return null;

    /* else ... */

    /* Building:  +- "[.]"
     *            |
     */

    ScrollTextOfDataComponent  jumpText
				     = new ScrollTextOfDataComponent ( this,
				  DataJumpStatement.PARSE_GUI_JUMP_STATEMENT );

    GraphNode       jumpNode = new GraphNode ( jumpText,
					       GraphComponent.LEFT,
					       DataJumpStatement.JUMP_COLOR,
					       true, null );
						     
    GraphConnection toJump = new GraphConnection( GraphComponent.LEFT,
						  DataJumpStatement.JUMP_COLOR,
						  false, false, jumpNode );

    GraphMeet       startMeet = new GraphMeet ( GraphComponent.UP,
						DataJumpStatement.JUMP_COLOR,
						GraphComponent.LEFT,
						toJump );

    GraphConnection tail = new GraphConnection ( GraphComponent.UP,
					     DataComponent.getDraggableColor(),
					     true, true, null );
    startMeet . addChild ( tail );

    return new GraphModule ( startMeet, tail, null, null,
			     this, true, theGraphType );
  }

	/* Invoked by this.update(*) invoking GraphModule.update() */
  public boolean updateGraphModule ( DataComponent  theModifiedComponent,
				     Object         theStoredData,
				     GraphModule    theGraphModule )
  {
	/* There is nothing to do here.. */
    return true;
  }
}
