
import java.io.PrintStream;
import netscape.application.*;
import netscape.util.*;

public class GraphMeet extends GraphComponent
{
  public final static int  DEFAULT_WIDTH  = 9;
  public final static int  DEFAULT_HEIGHT = 9;


  protected int             openCloseBranches;
  protected boolean         isOpen, isDisplayingInverse;
  protected GraphComponent  hiddenChildren[];


  public GraphMeet ( )
  {
    super ( 3 ); /* We can have between 0 and 2 children, inclusive */

    openCloseBranches = GraphComponent.NONE;
    isOpen = true;
    isDisplayingInverse = false;

    hiddenChildren = new GraphComponent[3];
    for ( int i=0;  i < hiddenChildren.length;  i++ )
      hiddenChildren[i] = null;
  }

	/* Convenience initializer */
  public GraphMeet ( int             theParentDirection,
		     Color           theForegroundColor,
		     int             theOpenOrCloseBranches,
		     GraphComponent  theChild )
  {
    this();
    setParentDirection   ( theParentDirection     );
    setOpenCloseBranches ( theOpenOrCloseBranches );
    setForegroundColor   ( theForegroundColor     );
    if ( theChild != null )
      addChild ( theChild );
  }


  public boolean getIsOpen        ( ) { return isOpen == true;  }
  public boolean getIsClosed      ( ) { return isOpen == false; }

  public void setIsOpen ( )
  {
	/* Idiocy check */
    if ( getIsOpen() == true )
      return;

	/* Record our state... */
    isOpen = true;

	/* Open up... */
    for ( int i=0;  i < getHiddenChildren().length;  i++ )
    {
      if ( getHiddenChildren()[i] != null )
      {
	getHiddenChildren()[i] . setParent ( null, false, false, false );
	super . addChild ( getHiddenChildren()[i] );
	getHiddenChildren()[i] = null;
      }
    }

    didRealign();  /* To Realign our children */
  }

  public void setIsClosed ( )
  {
    GraphModule     ourGraphModule = getOwningGraphModule();     
    GraphComponent  oldParentOfChild;
    int             direction, childIndex, hiddenChildrenIndex=0;

	/* Idiocy check */
    if ( getIsClosed() == true )
      return;

	/* Record our state... */
    isOpen = false;


	/* Close up... */
    for ( int i=0;  i < 4; i++ )
    {
	/* Find our direction */
      switch ( i )
      {
	case 0:  direction = GraphComponent.LEFT;  break;
	case 1:  direction = GraphComponent.RIGHT;  break;
	case 2:  direction = GraphComponent.DOWN;  break;
	case 3:  direction = GraphComponent.UP;  break;
	default:
	  System.err.println ( "[GraphMeet:setIsClosed]  "
			       + "Programmer Error Encountered.  Failing" );
	  return;
      }

		/* Are we closing this direction? */
      if ( GraphComponent.testDirection ( getOpenCloseBranches(), direction ) )
      {
		/* Do we have a child in this direction? */
	childIndex = findChildIndex ( direction );
	if ( childIndex != GraphComponent.INVALID )
	{
		/* Make our closed part unselected. */
	  if ( ourGraphModule == null )
	    getChild ( childIndex ) . setUnselected ( true, false );
	  else
	    ourGraphModule . setUnselected ( getChild ( childIndex ) );


		/* Find our next empty hidden-child slot */
	  for ( ;
		(   ( hiddenChildrenIndex < getHiddenChildren().length    )
		 && ( getHiddenChildren() [ hiddenChildrenIndex ] != null ) ) ;
		hiddenChildrenIndex ++ )
	    ; /* EMPTY FOR BODY */
		  
		/* Head off any catastrophic failures */
	  if ( hiddenChildrenIndex >= getHiddenChildren().length )
	  {
	    System.err.println ( "[GraphMeet:setIsClosed]  Error: "
				 + "Hiding too many children!!!" );
	    return;
	  }

	  getHiddenChildren()[ hiddenChildrenIndex ]
	    = getChild ( childIndex );

	  oldParentOfChild
	    = getHiddenChildren()[ hiddenChildrenIndex ] . getParent();

	  removeChild ( childIndex );

	  getHiddenChildren()[ hiddenChildrenIndex ]
	    . setParent ( oldParentOfChild, false, false, false );

	} /* If ( we have a child in this direction ) */

      } /* If ( we are closing this direction ) */

    } /* For ( each direction ) */

    didRealign();  /* To Realign our children */
  }


  public void toggleOpenClosed ( )
  {
    if ( getIsOpen() )
      setIsClosed();
    else
      setIsOpen();
  }

  public boolean getIsDisplayingInverse ( )
				    { return isDisplayingInverse == true;  }
  public boolean getIsDisplayingNormal  ( )
				    { return isDisplayingInverse == false; }

  public void setDisplayInverse ( )
  {
    if ( getIsDisplayingNormal() )
    {
      isDisplayingInverse = true;
      redisplay ( getComponentBounds() );
    }
  }

  public void setDisplayNormal ( )
  {
    if ( getIsDisplayingInverse() )
    {
      isDisplayingInverse = false;
      redisplay ( getComponentBounds() );
    }
  }


  public int  getOpenCloseBranches ( ) { return openCloseBranches; }

  public void setOpenCloseBranches ( int theBranches )
  {
    boolean wasClosed = getIsClosed();
    setIsOpen();
    openCloseBranches = theBranches;
    if ( wasClosed )
      setIsClosed();
  }


  protected GraphComponent[] getHiddenChildren() { return hiddenChildren; }


	/* addChild is overriden to handle the case where we try to
	 * add a child in the same direction as a hidden (closed) child.
	 */
  public boolean addChild ( GraphComponent theComponent )
  {
    for ( int i=0;  i < getHiddenChildren().length;  i++ )
    {
      if (   ( getHiddenChildren()[i] != null )
	  && ( GraphComponent . testDirection ( 
				getHiddenChildren()[i] . getParentDirection(),
				theComponent           . getParentDirection() )
	      ) )
      {
	System.err.println ( "[GraphMeet:addChild]  Error: "
			     + "Direction already taken by hidden child.  ("
			     + theComponent . getParentDirection() + ")" );
	return false;
      }
    }

    return super . addChild ( theComponent );
  }




  public void realign ( )
  {
    setWidth  ( GraphMeet.DEFAULT_WIDTH  );
    setHeight ( GraphMeet.DEFAULT_HEIGHT );

    super . realign();
  }



    /** Paints this object and all of its children. */
  public void drawSelf ( Graphics theGraphics )
  {
    GraphComponent   child;
    int              whichChildren, maxChildren, i,
                     xmin = minX(),
                     ymin = minY(),
                     xmax = maxX(),
                     ymax = maxY();


	/* If we are drawing in inverse mode... */
    if ( getIsDisplayingInverse() )
    {
      theGraphics . fillRect ( getComponentBounds() );
      theGraphics . setColor ( getBackgroundColor() );
      xmin += 1;
      ymin += 1;
      xmax -= 2;
      ymax -= 2;
    }

	/* Draw a circle */
    theGraphics . drawOval ( xmin, ymin, xmax - xmin, ymax - ymin );

	/* Draw line to parent */
    if ( getParent() != null )
    {      
      switch ( getParentDirection() )
      {
	case GraphComponent.RIGHT:
	  theGraphics . drawLine ( midX(), midY(),  xmax,  midY() );
	  break;

	case GraphComponent.LEFT:
	  theGraphics . drawLine ( midX(), midY(),  xmin,  midY() );
	  break;

	case GraphComponent.DOWN:
	  theGraphics . drawLine ( midX(), midY(), midX(),  ymax  );
	  break;

	case GraphComponent.UP:
	  theGraphics . drawLine ( midX(), midY(), midX(),  ymin  );
	  break;

	default:
	  System.err.println ( "[GraphMeet:drawSelf]  Warning:  "
			       + "Illegal value for getParentDirection().  ("
			       + getParentDirection() + ")" );
	  break;
      }
    }

	/* Draw line to children */
    for ( whichChildren=0;  whichChildren<2; whichChildren++ )
    {
      if ( whichChildren == 0 )
	maxChildren = getMaxChildren();
      else
	maxChildren = getHiddenChildren().length;


      for ( i=0; i < maxChildren; i++ )
      {
	if ( whichChildren == 0 )
	  child = getChild ( i );
	else
	  child = getHiddenChildren()[i];

	if ( child != null )
	{
	  switch ( child . getParentDirection() )
	  {
	    case GraphComponent.RIGHT:
	      theGraphics.drawLine ( midX(), midY(),  xmin,  midY() );
	      break;

	    case GraphComponent.LEFT:
	      theGraphics.drawLine ( midX(), midY(),  xmax,  midY() );
	      break;

	    case GraphComponent.DOWN:
	      theGraphics.drawLine ( midX(), midY(), midX(),  ymin  );
	      break;

	    case GraphComponent.UP:
	      theGraphics.drawLine ( midX(), midY(), midX(),  ymax  );
	      break;
	  } /* switch ( child . getParentDirection() */
	} /* if ( child != null ) */
      } /* for ( i in children ) */
    } /* for ( which set of children to work on ) */
  }


  public GraphComponent mouseDown ( MouseEvent theEvent )
  {
    if ( isRightMouseButtonDown ( theEvent ) == false )
      return super . mouseDown ( theEvent );

    setDisplayInverse();
    return this;
  }

  public void mouseDragged ( MouseEvent theEvent )
  {
    if ( getComponentBounds() . contains ( theEvent.x, theEvent.y ) )
      setDisplayInverse();
    else
      setDisplayNormal();
  }

  public void mouseUp ( MouseEvent theEvent )
  {
    if ( getIsDisplayingInverse() )
    {
      toggleOpenClosed();
      setDisplayNormal();
    }
  }


}

