
import java.io.PrintStream;
import netscape.application.*;
import netscape.util.*;

public class GraphNode extends GraphComponent
{
  public final static int  DEFAULT_BORDER_SIZE = 3;
  public final static int  ABSOLUTE_MINIMUM_SIZE = 10;

  protected int        borderSize;
  protected View       containedView;
  protected GraphRoot  rootNode;
  protected int        mouseDownX, mouseDownY, mouseQuadrant;


  public GraphNode ( View theContainedView )
  {
    super ( 3 ); /* We have 0 to 3 children. */
    borderSize = GraphNode . DEFAULT_BORDER_SIZE;
    containedView = null;
    rootNode = null;
    mouseDownX = 0;
    mouseDownY = 0;
    mouseQuadrant = GraphComponent.ILLEGAL;
    setContainedView ( theContainedView );
  }

	/* Convenience initializer */
  public GraphNode ( View            theContainedView,
		     int             theParentDirection,
		     Color           theForegroundColor,
		     boolean         theIsResizable,
		     GraphComponent  theChild )
  {
    this ( theContainedView );
    setParentDirection   ( theParentDirection );
    setForegroundColor   ( theForegroundColor );
    setIsResizable       ( theIsResizable     );
    if ( theChild != null )
      addChild ( theChild );
  }


  public View  getContainedView ( )  { return containedView; }
  public void  setContainedView ( View theContainedView )
  {
    if ( getContainedView() != null )
      removeSubviewsUp ( this );
    containedView = theContainedView;
    addSubviewsUp ( this );
    realign();
  }

  public int  getBorderSize ( ) { return borderSize; }
  public void setBorderSize ( int theBorderSize )
				{ borderSize = theBorderSize; }

  public GraphRoot getRoot (                   ) { return rootNode;    }
  public void      setRoot ( GraphRoot theRoot ) { rootNode = theRoot; }

  public boolean getIsResizable ( )
  {
    return mouseQuadrant != GraphComponent.ILLEGAL;
  }
  public void setIsResizable ( boolean theIsResizable )
  {
    if ( theIsResizable == false )
      mouseQuadrant = GraphComponent.ILLEGAL;
    else if ( mouseQuadrant == GraphComponent.ILLEGAL )
      mouseQuadrant = GraphComponent.NONE;
  }


  public void realign ( )
  {
    int x, y;

    setWidth  ( getContainedView() . bounds() . width 
	       +  (2 * getBorderSize()) );
    setHeight ( getContainedView() . bounds() . height
	       +  (2 * getBorderSize()) );
    super.realign();

    x = getX() + getBorderSize();
    y = getY() + getBorderSize();

    if ( getRoot() != null )
    {
      x += getRoot() . getSubviewXAdjust();
      y += getRoot() . getSubviewYAdjust();
    }

    getContainedView() . moveTo ( x, y );
  }


  public void drawSelf ( Graphics theGraphics )
  {
    for ( int i=0; i < getBorderSize(); i++ )
    {
      theGraphics . drawRect ( getComponentBounds().x + i,
			       getComponentBounds().y + i,
			       getComponentBounds().width  - (2 * i),
			       getComponentBounds().height - (2 * i) );
    }
  }


  public void addSubviewsDown ( GraphRoot theRoot )
  {
	/* Cache theroot for use in realign... */
    setRoot ( theRoot );

    if ( getContainedView() != null )
    {
      theRoot . addSubview ( this, getContainedView() );
    }
    super . addSubviewsDown ( theRoot );
  }


  public void removeSubviewsDown ( GraphRoot theRoot )
  {
    if ( theRoot != getRoot() )
    {
      System.err.println ( "[GraphNode:removeSubviews]  Warning:  "
			   + "theRoot != getRoot().  "
			   + "Internal consistency failure." );
    }

    if ( getContainedView() != null )
    {
      theRoot . removeSubview ( this, getContainedView() );
    }
    super . removeSubviewsDown ( theRoot );
  }


  public boolean destroySubtree()
  {
    boolean  returnValue = true;

    if (   ( getContainedView() != null                          )
	&& ( getContainedView() instanceof DataComponentObserver )
	&& ( ( (DataComponentObserver) getContainedView() )
	       . unregisterToObserve() == false                  ) )
      returnValue = false;

    if ( super . destroySubtree() == false )
      returnValue = false;

    return returnValue;
  }



  public void print ( PrintStream theStream, boolean doRecursive )
  {
    theStream . println ( "containedView....: "
			  + ( getContainedView() == null ? "null"
			      : getContainedView() . getClass() . getName() ));
	/* This method is just for debugging, so I'll kluge "contents" here. */
    if ( getContainedView() instanceof ScrollText )
      theStream . println ( "...ScrollText....: \""
			    + ( ( (ScrollText) getContainedView() )
			       . getTextView() . string() ) + "\"" );
    else if ( getContainedView() instanceof SimpleLabel )
      theStream . println ( "...SimpleLabel...: \""
			    + ( ( (SimpleLabel) getContainedView() )
			       . getText() ) + "\"" );
    theStream . println ( "borderSize.......: " + getBorderSize() );
    super . print ( theStream, doRecursive );
  }



  public GraphComponent mouseDown ( MouseEvent theEvent )
  {
    if ( isRightMouseButtonDown ( theEvent ) == false )
      return super . mouseDown ( theEvent );

    if ( getIsResizable() == false )
      return null;

    mouseDownX    = theEvent.x;
    mouseDownY    = theEvent.y;

    if ( mouseDownX < ( getX() + ( getWidth() / 2 ) ) )
      mouseQuadrant = GraphComponent.LEFT;
    else
      mouseQuadrant = GraphComponent.RIGHT;

    if ( mouseDownY < ( getY() + ( getHeight() / 2 ) ) )
      mouseQuadrant = GraphComponent . addDirection ( mouseQuadrant,
						      GraphComponent.UP );
    else
      mouseQuadrant = GraphComponent . addDirection ( mouseQuadrant,
						      GraphComponent.DOWN );

    return this;
  }

  public void mouseDragged ( MouseEvent theEvent )
  {
    if ( getIsResizable() == false )
      return;

    if ( getContainedView() == null )
      return;

    int width  = getContainedView() . bounds() . width,
        height = getContainedView() . bounds() . height,
        widthFactor  = 1,
        heightFactor = 1;

	/* Eventually compensate for direction that we don't center */
        /* For the moment -- don't compensate.  It introduces a
	 * horrible scaling bug. */
/*    if ( GraphComponent . testDirection ( getParentDirection(),
					  GraphComponent.HORIZONTAL ) )
    {
      widthFactor = 1;
    }

    if ( GraphComponent . testDirection ( getParentDirection(),
					  GraphComponent.VERTICAL ) )
    {
      heightFactor = 1;
    }
*/
    if ( GraphComponent . testDirection ( mouseQuadrant, GraphComponent.LEFT) )
    {
      width += widthFactor * ( mouseDownX - theEvent.x );
    }

    if ( GraphComponent . testDirection ( mouseQuadrant, GraphComponent.RIGHT))
    {
      width -= widthFactor * ( mouseDownX - theEvent.x );
    }

    if ( GraphComponent . testDirection ( mouseQuadrant, GraphComponent.UP ) )
    {
      height += heightFactor * ( mouseDownY - theEvent.y );
    }

    if ( GraphComponent . testDirection ( mouseQuadrant, GraphComponent.DOWN) )
    {
      height -= heightFactor * ( mouseDownY - theEvent.y );
    }

    if ( width < GraphNode . ABSOLUTE_MINIMUM_SIZE )
      width = GraphNode . ABSOLUTE_MINIMUM_SIZE;

    if ( height < GraphNode . ABSOLUTE_MINIMUM_SIZE )
      height = GraphNode . ABSOLUTE_MINIMUM_SIZE;

    getContainedView() . sizeTo ( width, height );

    mouseDownX    = theEvent.x;
    mouseDownY    = theEvent.y;
  }

  public void mouseUp ( MouseEvent theEvent )
  {
    mouseDownX = 0;
    mouseDownY = 0;
    if ( getIsResizable() )
      mouseQuadrant = GraphComponent.NONE;
  }

  public void subviewDidResize ( View theSubview )
  {
    realign();
  }
}
