
import java.io.PrintStream;
import netscape.application.*;
import netscape.util.*;

/** The Root of any graph.  Interfaces the graph with the GraphView.
  * Determines the type of the graph. 
  */

public class GraphRoot extends GraphComponent
{
	/* Define the known graph types */
  public final static int ILLEGAL_GRAPH  = 0;  /* Note: these are bit-flags! */
  public final static int GUI_TREE_GRAPH = 1;
  public final static int _2D_TREE_GRAPH = 2;

    /** Class method to test graph types */
  public static boolean testGraphType ( int theGraphType, int theTestCriteria )
  {
    return ( ( theGraphType & theTestCriteria ) != 0 );
  }
    /** Class method to add two graph types together */
  public static int addGraphType ( int theGraphType, int theOtherGraphType )
  {
    return theGraphType | theOtherGraphType;
  }



  protected int        graphType;
  protected GraphView  parentView;
  protected Hashtable  graphComponentForView;

  public GraphRoot ( )
  {
    super ( 4 );
    graphType = GraphRoot.GUI_TREE_GRAPH;
    setParentView ( null );
    graphComponentForView = new Hashtable();
    setX      ( 10 );
    setY      ( 10 );
    setWidth  ( 1 );
    setHeight ( 1 );
  }

  public int  getGraphType (                  ) { return graphType;         }
  public void setGraphType ( int theGraphType ) { graphType = theGraphType; }

  public GraphView getParentView ( ) { return parentView; }

  public void setParentView ( GraphView theParentView )
  {
    if ( getParentView() != null )
      removeSubviewsUp ( this );
    parentView = theParentView;
    addSubviewsUp ( this );
  }

  public void setParent ( GraphComponent theParent )
  {
    System.err.println ( "[GraphRoot:setParent]  Error:  "
			 + "Called setparent on a GraphRoot object." );
  }

  protected GraphComponent getGraphComponentForView ( View theView )
  {
    GraphComponent result
      = (GraphComponent ) graphComponentForView . get ( theView );
    return result;
  }

  protected GraphComponent setGraphComponentForView (
						 GraphComponent theComponent,
						 View           theView )
  {
    GraphComponent result
      = (GraphComponent) graphComponentForView . put ( theView, theComponent );

    if ( result != null )
    {
      System.err.println ( "[GraphRoot:setGraphComponentForView]  Warning:  "
			   + "Prior GraphComponent was set for View. " );
      result . print ( System.err );
    }
    return result;
  }

  protected GraphComponent clearGraphComponentForView ( View theView )
  {
    GraphComponent result
      = (GraphComponent) graphComponentForView . remove ( theView );

    if ( result == null )
      System.err.println ( "[GraphRoot:clearGraphComponentForView]  Warning:  "
			   + "No GraphComponent for View: " + theView );
    return result;
  }


  public void drawSelf ( Graphics theGraphics )
  {
	/* Nothing to draw */
  }

  public void didRealignChildren ( GraphComponent theRealignedChildTree,
				   GraphComponent theOriginalChild      )
  {
    super . didRealignChildren ( theRealignedChildTree, theOriginalChild );

    if ( getParentView() != null )
    {
      getParentView() . resizeTo ( getSubgraphBounds(),
				   (  ( theOriginalChild == null )
				     ? null
				     : theOriginalChild . getComponentBounds()
				    ) );
    }
  }

  public void redisplay ( Rect theRectToRedisplay )
  {
    if ( getParentView() != null )
    {
	/* BUGFIX:  Linux console doesn't like it if we addDirtyRect(NULL) */
      if ( theRectToRedisplay == null )
	theRectToRedisplay = getSubgraphBounds();

      theRectToRedisplay . x += getSubviewXAdjust();
      theRectToRedisplay . y += getSubviewYAdjust();
      getParentView() . addDirtyRect ( theRectToRedisplay );
      theRectToRedisplay . x -= getSubviewXAdjust();
      theRectToRedisplay . y -= getSubviewYAdjust();
    }
    else
      System.err.println ( "[GraphRoot:redisplay]  Error:  No ParentView." );
  }


  public void addSubviewsUp ( GraphComponent theComponent )
  {
    if ( getParentView() != null )
      theComponent . addSubviewsDown ( this );
  }

  public void addSubview ( GraphComponent theComponent, View theView )
  {
    getParentView() . addSubview ( theView );
    setGraphComponentForView ( theComponent, theView );
  }

  public void removeSubviewsUp ( GraphComponent theComponent )
  {
    if ( getParentView() != null )
      theComponent . removeSubviewsDown ( this );
  }

  public void removeSubview ( GraphComponent theComponent, View theView )
  {
    theView . removeFromSuperview();
    clearGraphComponentForView ( theView );
  }

  public int getSubviewXAdjust()
  {
    if ( getParentView() != null )
      return getParentView() . getSubviewXAdjust();
    else
      return 0;
  }

  public int getSubviewYAdjust()
  {
    if ( getParentView() != null )
      return getParentView() . getSubviewYAdjust();
    else
      return 0;
  }


  public void subviewDidResize ( View theSubview )
  {
    GraphComponent owner = getGraphComponentForView ( theSubview );
    if ( owner != null )
      owner . subviewDidResize ( theSubview );
  }


  public void debugprintviews( View theView, String title )
  {
    System.out.println("\nTitle.........: " + title );
    System.out.println("  View........: " + theView . getClass() . getName() );
    System.out.println("  IsBuffered..: " + theView . isBuffered() );
    System.out.println("  IsTranspar..: " + theView . isTransparent() );
    Vector subviewsVector = theView.subviews();
    int i;
    for (i=0; i<subviewsVector.count(); i++ )
      debugprintviews ( (View) subviewsVector.elementAt(i), title + " - " + i );
  }

  public GraphComponent mouseDown ( MouseEvent theEvent )
  {
    GraphComponent targetComponent = findComponent ( theEvent.x , theEvent.y );

	/* If event is inside me, or is not inside my children... */
    if (   ( targetComponent == this )
        || ( targetComponent == null ) )
    {
      return null;
    }
    else
      return targetComponent . mouseDown ( theEvent );
  }

	/* DragSource Interface */
  public View sourceView ( DragSession theDragSession )
  {
    return getParentView();
  }

}
