
import netscape.application.*;
import netscape.util.*;

public class DefaultTextFilter extends Object implements TextFilter
{
	/* Class Constants */
  public static final String CUT_COMMAND           = ExtendedTarget.CUT;
  public static final String COPY_COMMAND          = ExtendedTarget.COPY;
  public static final String PASTE_COMMAND         = ExtendedTarget.PASTE;
  public static final String DELETE_COMMAND        = "DELETE";
  public static final String PREVIOUS_VIEW_COMMAND = "SelectPreviousView";
  public static final String NEXT_VIEW_COMMAND     = "SelectNextView";
  
  public static final int    CUT_KEY           = 'x';
  public static final int    COPY_KEY          = 'c';
  public static final int    PASTE_KEY         = 'v';
  public static final int    DELETE_KEY        = 'd';
  public static final int    PREVIOUS_VIEW_KEY = 'i'; /* with wonky modifiers*/
  public static final int    NEXT_VIEW_KEY     = 'i'; /* with wonky modifiers*/

  public static int actualValueForKey ( int theKey ) {return theKey - 'a' + 1;}


	/* The [singleton] default text filter */
  public static DefaultTextFilter  defaultTextFilter = new DefaultTextFilter();


	/* Setup methods */
  public static void setDefaultTextFilter ( View theView )
  {
    if ( theView instanceof TextField )
    {
      setDefaultTextFilter ( (TextField) theView );
      return;
    }

    if ( theView instanceof TextView )
    {
      setDefaultTextFilter ( (TextView) theView );
      return;
    }

    Vector subviews = theView . subviews();
    int    i;

    for ( i=0;  i < subviews.count();  i++ )
    {
      if      ( subviews . elementAt ( i ) instanceof TextField )
	setDefaultTextFilter ( (TextField) subviews . elementAt ( i ) );

      else if ( subviews . elementAt ( i ) instanceof TextView )
	setDefaultTextFilter ( (TextView) subviews . elementAt ( i ) );

      else if ( subviews . elementAt ( i ) instanceof View )
	setDefaultTextFilter ( (View) subviews . elementAt ( i ) );
    }
  }

  public static void setDefaultTextFilter ( TextField theTextField )
  {
    theTextField . setFilter ( DefaultTextFilter.defaultTextFilter );
  }

  public static void setDefaultTextFilter ( TextView theTextView )
  {
    theTextView . setFilter ( DefaultTextFilter.defaultTextFilter );
  }



	/* TextFilter Interface */
  public boolean acceptsEvent ( Object   theTextObject,
				KeyEvent theEvent,
				Vector   theEvents )
  {
    switch ( theEvent.key )
    {
      case 0x01: /* CTRL-A */   theEvent.key = KeyEvent.HOME_KEY;        break;
      case 0x05: /* CTRL-E */   theEvent.key = KeyEvent.END_KEY;         break;
      case 0x02: /* CTRL-B */   theEvent.key = KeyEvent.LEFT_ARROW_KEY;  break;
      case 0x06: /* CTRL-F */   theEvent.key = KeyEvent.RIGHT_ARROW_KEY; break;
      case 0x10: /* CTRL-P */   theEvent.key = KeyEvent.UP_ARROW_KEY;    break;
      case 0x0E: /* CTRL-N */   theEvent.key = KeyEvent.DOWN_ARROW_KEY;  break;
      case 0x04: /* CTRL-D */   theEvent.key = KeyEvent.DELETE_KEY;      break;
      case KeyEvent.DELETE_KEY: theEvent.key = KeyEvent.BACKSPACE_KEY;   break;
    }

	/* These should almost always have already been properly handled. */
	/* But just in case... */
    if ( theTextObject instanceof ExtendedTarget )
    {
      if (   ( theEvent.key == actualValueForKey( DefaultTextFilter.CUT_KEY ) )
	  && ( ((ExtendedTarget) theTextObject)
	       . canPerformCommand ( DefaultTextFilter.CUT_COMMAND ) ) )
      {
	Application . application() . performCommandLater (
						 ((Target) theTextObject),
						 DefaultTextFilter.CUT_COMMAND,
						 null, false );
	return false;
      }


      if (   ( theEvent.key == actualValueForKey( DefaultTextFilter.COPY_KEY ))
	  && ( ((ExtendedTarget) theTextObject)
	       . canPerformCommand ( DefaultTextFilter.COPY_COMMAND ) ) )
      {
	Application . application() . performCommandLater (
						((Target) theTextObject),
						DefaultTextFilter.COPY_COMMAND,
						null, false );
	return false;
      }


      if (   ( theEvent.key == actualValueForKey( DefaultTextFilter.PASTE_KEY))
	  && ( ((ExtendedTarget) theTextObject)
	       . canPerformCommand ( DefaultTextFilter.PASTE_COMMAND ) ) )
      {
	Application . application() . performCommandLater (
					       ((Target) theTextObject),
					       DefaultTextFilter.PASTE_COMMAND,
					       null, false );
	return false;
      }


      if (   ( theEvent.key == actualValueForKey(DefaultTextFilter.DELETE_KEY))
	  && ( ((ExtendedTarget) theTextObject)
	       . canPerformCommand ( DefaultTextFilter.DELETE_COMMAND ) ) )
      {
	Application . application() . performCommandLater (
					      ((Target) theTextObject),
					      DefaultTextFilter.DELETE_COMMAND,
					      null, false );
	return false;
      }


      if (   (    theEvent.key
	       == actualValueForKey ( DefaultTextFilter.PREVIOUS_VIEW_KEY ) )
	  && ( ((ExtendedTarget) theTextObject)
	       . canPerformCommand ( DefaultTextFilter.PREVIOUS_VIEW_COMMAND) )
	  && ( theEvent . isShiftKeyDown() == true ) )
      {
	Application . application() . performCommandLater (
				       ((Target) theTextObject),
				       DefaultTextFilter.PREVIOUS_VIEW_COMMAND,
				       null, false );
	return false;
      }


      if (   (   theEvent.key
	      == actualValueForKey ( DefaultTextFilter.NEXT_VIEW_KEY ) )
	  && ( ((ExtendedTarget) theTextObject)
	       . canPerformCommand ( DefaultTextFilter.NEXT_VIEW_COMMAND ) )
	  && ( theEvent . isShiftKeyDown() == false ) )
      {
	Application . application() . performCommandLater (
					   ((Target) theTextObject),
					   DefaultTextFilter.NEXT_VIEW_COMMAND,
					   null, false );
	return false;
      }
    }

    return true;
  }

}
