
import netscape.application.*;
import netscape.util.*;
import netscape.constructor.*;


/*Abstract class*/
public class FileTaskWindowHandler extends WindowHandlerAbstractBaseClass
			        implements ExtendedTarget, Target
{
	/* TASK_WINDOW_PLAN Objects */
  public final static String FILE_LIST_VIEW          = "FileListView";
  public final static String TASK_LIST_VIEW          = "TaskListView";

	/* Commands */
  public final static String SET_FILE_COMMAND        = "SetFile";
  public final static String NULL_COMMAND            = "NullCommand";


	/* Instance Variables */
  protected ListView   fileListView,
                       taskListView;
  protected ListItem   allFilesListItem;


	/* Constructors */
  public FileTaskWindowHandler ()
  {
    fileListView     = null;
    taskListView     = null;
    allFilesListItem = null;
  }

	/* Constructors can't be synchronized. */
	/* And this needs to be synchronized since we use the static plans */
  protected synchronized
      boolean createAndInitialize ( Plan     thePlan,
				    String   theFileDoubleClickCommand,
				    String   theTaskDoubleClickCommand,
				    boolean  theUseExternalWindow,
				    boolean  theHasAllFilesListItem,
				    String   theWindowTitle,
				    String   theErrorMessageLocation,
				    VDT      theVDT )
  {
    if ( super . createAndInitialize ( thePlan,
				       theUseExternalWindow,
				       theWindowTitle,
				       theErrorMessageLocation,
				       theVDT )
	 == false )
    {
      return false;
    }

	/* Lets find our instance variables */
    fileListView
      = (ListView)
         ( ( (ScrollGroup)
	     (thePlan.componentNamed ( FileTaskWindowHandler . FILE_LIST_VIEW )
	      ) ) . contentView() );
    taskListView
      = (ListView)
         ( ( (ScrollGroup)
	     (thePlan.componentNamed ( FileTaskWindowHandler . TASK_LIST_VIEW )
	      ) ) . contentView() );

	/* Error-check */
    if (   ( fileListView         == null )
	|| ( taskListView         == null ) )
    {
      System.err.println ( "[FileTaskWindowHandler:createAndInitialize]  "
	   + "Warning:  One or more Plan objects is NULL!!!\n"
	   + "    fileListView...........: " + fileListView         + "\n"
	   + "    taskListView...........: " + taskListView         + "\n"
	   + "------------------------------------------------------------" );
    }


	/* Perform object reconfigurations... */
    if ( theHasAllFilesListItem )
      allFilesListItem = fileListView . itemAt ( 0 );
    else
      allFilesListItem = null;

    if ( theFileDoubleClickCommand != null )
      fileListView . setDoubleCommand ( theFileDoubleClickCommand );
    /* else -- fileListView already has a single-click command to do... */

    taskListView . setTarget        ( this );
    taskListView . setCommand       ( FileTaskWindowHandler.NULL_COMMAND );

    if ( theTaskDoubleClickCommand != null )
      taskListView . setDoubleCommand ( theTaskDoubleClickCommand );
    else
      taskListView . setDoubleCommand ( FileTaskWindowHandler.NULL_COMMAND );



	/* Report that we succeeded. */
    return true;
  }


  public void runModal ( )
  {
	/* Recreate the files list. */
    loadKnownFiles();

	/* Recreate the tasks list. */
    performSetFile();

    super . runModal();
  }

  public void run ( )
  {
	/* Recreate the files list. */
    loadKnownFiles();

	/* Recreate the tasks list. */
    performSetFile();

    super . run();
  }



  public RootView    getRootView()   { return getFileListView() . rootView(); }
  public ListView    getFileListView()        { return fileListView;         }
  public ListView    getTaskListView()        { return taskListView;         }
  public ListItem    getAllFilesListItem()    { return allFilesListItem;     }


	/* ExtendedTarget Interface */
  public boolean canPerformCommand ( String theCommand ) 
  {
    return theCommand . equals ( FileTaskWindowHandler.SET_FILE_COMMAND );
  }


	/* Target Interface */
  public void performCommand ( String theCommand, Object theData )
  {
	/* Handle case of adding more arguments */
    if ( theCommand . equals ( FileTaskWindowHandler.SET_FILE_COMMAND ) )
    {
      performSetFile();
    }

    else if ( theCommand . equals ( FileTaskWindowHandler.NULL_COMMAND ) )
    {
	/* This is a no-op to deal with single-click command coming from */
	/* the Tasks ListView when the first mouse-click occurs...       */
      ;
    }

    else
    {
      super . performCommand ( theCommand, theData );
    }

  } /* public void performCommand ( String theCommand, Object theData ) */



  public void loadKnownFiles ( )
  {
	/* Clear out File's List View */
    getFileListView() . removeAllItems();

	/* Add back our select all files option... */
	/* (If we have one...) */
    if ( getAllFilesListItem() != null )
      getFileListView() . addItem ( getAllFilesListItem() );

	/* Add known files... */
    Enumeration  knownFiles = DataFile.getRegisteredFiles();
    DataFile     dataFile;
    ListItem     listItem;

    while ( knownFiles . hasMoreElements() )
    {
      dataFile = (DataFile) knownFiles.nextElement();
      listItem = getFileListView() . addItem();
      listItem . setTitle ( dataFile.getFilename() );
      listItem . setData  ( dataFile );
    }
	/* Select all-files option, if we have one... */
    if ( getAllFilesListItem() != null )
      getFileListView() . selectItem ( getAllFilesListItem() );
    else if ( getFileListView() . count() > 0 )
      getFileListView() . selectItemAt ( 0 );
    /* else -- nothing there to select... */

	/* And resize... */
    getFileListView() . sizeToMinSize();
  }


  public void performSetFile ( )
  {
    DataFile            dataFile;
    DataTaskDefinition  task;
    ListItem            listItem;

	/* Clear out the Task List View */
    getTaskListView() . removeAllItems();

	/* Are we doing all the files, or just one file? */
    if (   ( getFileListView() . selectedItem() == getAllFilesListItem() )
	&& ( getAllFilesListItem()              != null                  ) )
    {
      Enumeration         knownTasks = DataTaskDefinition.getRegisteredTasks();

      while ( knownTasks . hasMoreElements() )
      {
	task     = (DataTaskDefinition) knownTasks.nextElement();
	listItem = getTaskListView() . addItem();
	listItem . setTitle ( task.getTaskName() );
	listItem . setData  ( task );
      }
    }

    else if (   ( getFileListView() . selectedItem() != null )
	     && ( getFileListView() . count()         > 0    ) )
    {
      dataFile = (DataFile) ( getFileListView() . selectedItem() . data() );

      for ( int i=0;  i < dataFile.getSubcomponentsCount();  i++ )
      {
	if (   ( dataFile.isSubcomponentADataComponent ( i ) )
	    && ( dataFile.getDataComponentSubcomponent ( i )
		 instanceof DataTaskDefinition               ) )
	{
	  task = ( (DataTaskDefinition)
		   ( dataFile . getDataComponentSubcomponent ( i ) ) );
	  
	  listItem = getTaskListView() . addItem();
	  listItem . setTitle ( task.getTaskName() );
	  listItem . setData  ( task );
	}
      }
    }

    else
    {
	/* No file was selected... */
    }

	/* Try to select the first item, if we have one... */
    if ( getTaskListView().count() > 0 )
      getTaskListView().selectItemAt ( 0 );

	/* And resize... */
    getTaskListView() . sizeToMinSize();
    getTaskListView() . setDirty ( true );
  }


	/** Selects theDataFile, if it can find it...
	  * Otherwise does nothing...
	  */
  public boolean selectFile ( DataFile  theDataFile )
  {
    for ( int i=0;  i < getFileListView() . count();  i ++ )
    {
      if ( getFileListView() . itemAt ( i ) . data() == theDataFile )
      {
	getFileListView() . selectItemAt ( i );
	performSetFile();
	return true;
      }
    }
    return false;
  }

}
