
import netscape.application.*;
import netscape.util.*;

public class ScrollTextOfDataComponent extends ScrollText
				    implements DataComponentObserver,
					       Target, ExtendedTarget
{
	/* Class Constants/Variables */
  public static final String ALERT_PARSE_EXCEPTION = "AlertParseException";
  public static final String ALERT_PARSE_FAILURE   = "AlertParseFailure";


	/* General reedit/revert alert.                */
	/*  Returns True for Reedit,  False for Revert */
  public static boolean showReeditRevertAlert ( String   theTitle,
						String   theMessage,
						RootView theRootView )
  {
    return ( UTIL.runAlert ( theTitle, theMessage,
			     "Revert", "Reedit", null, theRootView )
	     == Alert.SECOND_OPTION );
  }



	/* Instance Variables */
  protected int            dataComponentSubset;
  protected DataComponent  dataComponent;
  protected int            delayedAlertCount;


	/* Instance Methods */
  public ScrollTextOfDataComponent ( )
  {
    super();
    privateInitialize();
  }
  public ScrollTextOfDataComponent ( Rect theRect )
  {
    super ( theRect );
    privateInitialize();
  }
  public ScrollTextOfDataComponent ( int theX,     int theY,
				     int theWidth, int theHeight )
  {
    super ( theX, theY, theWidth, theHeight );
    privateInitialize();
  }

  private void privateInitialize()
  {
    dataComponentSubset = DataComponent.ENTIRE_OBJECT;
    dataComponent       = null;
    delayedAlertCount   = 0;

	/* Lets show the entire text by default...  For now... */
    setShowEntireText ( true );
  }

	/* Convenience Constructor(s) */
  public ScrollTextOfDataComponent ( DataComponent theDataComponent )
  {
    this();
    setDataComponent ( theDataComponent );
  }
  public ScrollTextOfDataComponent ( DataComponent theDataComponent,
				     int           theDataComponentSubset )
  {
    this();
    setDataComponentSubset ( theDataComponentSubset );
    setDataComponent       ( theDataComponent       );
  }


  public int  getDataComponentSubset()  { return dataComponentSubset; }
  public void setDataComponentSubset( int theDataComponentSubset )
  {
    dataComponentSubset = theDataComponentSubset;
	/* Use update method to set our initial text... */
    if ( getDataComponent() != null )
      this . update ( getDataComponent(), getDataComponent(), null );
  }


  public DataComponent getDataComponent ( ) { return dataComponent; }
  public boolean       setDataComponent ( DataComponent theDataComponent )
  {
    boolean  returnValue = true;

    if ( dataComponent != null )
    {
      if ( dataComponent . removeObserver ( this ) == false )
	returnValue = false;
    }

    dataComponent = theDataComponent;

    if ( dataComponent != null )
    {
      if ( dataComponent . addObserver ( this, null ) == false )
	returnValue = false;

	/* Use update method to set our initial text... */
      this . update ( dataComponent, dataComponent, null );
    }

    return returnValue;
  }

  public boolean unregisterToObserve ()
  {
    return setDataComponent ( null );
  }

  public void setDataComponent ( DataComponent theDataComponent,
				 int           theDataComponentSubset )
  {
    setDataComponent       ( null                   );
    setDataComponentSubset ( theDataComponentSubset );
    setDataComponent       ( theDataComponent       );
  }

  public void delayedAlert ( String theAlertType, Object theData )
  {
    delayedAlertCount++;
    Application . application()
                . performCommandLater ( this, theAlertType, theData, true );
  }

  public int getDelayedAlertCount ()  { return delayedAlertCount; }


	/* Overriden part of inherited TextViewOwner interface */
  public void textEditingDidEnd ( TextView theTextView )
  {
		/* Abort if we are not part of the app's view hierarchy... */
    if (   ( isInViewHierarchy() == false )
		/* Abort if we don't have a underlying DataComponent */
	|| ( getDataComponent() == null )
		/* Abort if we haven't changed this string... */
	|| ( getText() . equals ( getDataComponent()
				  . toString ( getDataComponentSubset() ) ) )
        )
    {
      super . textEditingDidEnd ( theTextView );
      return;
    }


    try
    {
      if ( getDataComponent() . parse ( getText(),
					getDataComponentSubset() ) == false )
      {
	delayedAlert ( ScrollTextOfDataComponent.ALERT_PARSE_FAILURE, null );
      }
    }
    catch ( DetailedParseException  theException )
    {
      delayedAlert ( ScrollTextOfDataComponent.ALERT_PARSE_EXCEPTION,
		     theException );
    }
  }


  public boolean canPerformCommand ( String theString )
  {
    return
      (   theString . equals( ScrollTextOfDataComponent.ALERT_PARSE_FAILURE   )
       || theString . equals( ScrollTextOfDataComponent.ALERT_PARSE_EXCEPTION )
       );
  }

  public void performCommand ( String theCommand, Object theData )
  {
    if ( theCommand . equals( ScrollTextOfDataComponent.ALERT_PARSE_FAILURE ) )
    {
	/* Decrement our count.. */
      if ( delayedAlertCount > 0 )
	delayedAlertCount--;

      if ( ScrollTextOfDataComponent . showReeditRevertAlert (
		 "Internal Error:", "Unable to Use Parsed Task Definition.",
		 rootView() ) )
      {
	getTextView() . delayedSelectRange ( getTextView().selectedRange() );
      }
      else
      {
	setText ( getDataComponent() . toString ( getDataComponentSubset() ) );
      }
    }

    else if ( theCommand
	        . equals ( ScrollTextOfDataComponent.ALERT_PARSE_EXCEPTION ) )
    {
	/* Decrement our count.. */
      if ( delayedAlertCount > 0 )
	delayedAlertCount--;

      DetailedParseException  theException = (DetailedParseException) theData;

      if ( ScrollTextOfDataComponent . showReeditRevertAlert (
			   "Parse Error:", theException . toClippedString(),
			   rootView() ) )
      {
		/* Try to find the range corresponding to our error token */
	Range  selectRange = null;

	if ( theException . getHasErrorLocation() )
	{
	  selectRange = new Range ( TDLParser . getLastDataSource()
				      . getPositionFromLineAndColumn (
					  theException.getErrorLineStart(),
					  theException.getErrorColumnStart() ),
				    theException . getErrorLength() );
	}
		/* If we failed... */
	if (   ( selectRange        == null )
	    || ( selectRange.index  == DataSource.UNKNOWN_POSITION )
	    || ( selectRange.length == DetailedParseException.UNKNOWN_VALUE))
	{
	  selectRange = getTextView() . selectedRange();
	}

		/* And select our chosen range... */
	getTextView() . delayedSelectRange ( selectRange );
      }
      else
      {
	setText ( getDataComponent() . toString ( getDataComponentSubset() ) );
      }

    }
    else
    {
      System.err.println ( "[ScrollTextOfDataComponent:performCommand]  "
		   + "Warning:  Unknown command \"" + theCommand + "\"." );
    }
  }


	/* DataComponentObserver Interface */
  public boolean update ( DataComponent  theObservedComponent,
			  DataComponent  theModifiedComponent,
			  Object         theStoredData        )
  {
    if ( theObservedComponent != getDataComponent() )
    {
      System.err.println ( "[ScrollTextOfDataComponent:update("
	   + theObservedComponent . getClass() . getName()
	   + "," + theModifiedComponent . getClass() . getName()
	   + "," + ( (theStoredData == null)
		     ? "null" : theStoredData . getClass() . getName() )
	   + ")]  Warning:  theObservedComponent != getDataComponent().  ["
	   + ( (getDataComponent() == null ) ? "null"
	       : getDataComponent() . getClass() . getName() )
	   + "]" );
    }

    if ( getDataComponent() != null )
      setText( getDataComponent().toString( getDataComponentSubset(), true ) );
    else
      setText ( "NULL" );

    return true;
  }
}
