
import netscape.application.*;
import netscape.util.*;
import netscape.constructor.*;


public class SelectFileWindowHandler extends FileTaskWindowHandler
{
	/* Constant (Special) Filenames */
  public final static String SELECT_FILE_WINDOW_PLAN_NAME
						 = "SelectFileWindow.planb";

	/* TASK_WINDOW_PLAN Objects */
  public final static String CANCEL_BUTTON       = "CancelButton";
  public final static String SELECT_BUTTON       = "SelectButton";
  public final static String NEW_FILE_BUTTON     = "NewFileButton";


	/* Commands */
  public final static String CANCEL_COMMAND      = "Cancel";
  public final static String SELECT_COMMAND      = "Select";
  public final static String NEW_FILE_COMMAND    = "NewFile";

	/* Constants */
  public final static String TITLE = "Select File";


	/* Class variables */
  protected static Plan                     selectFileWindowPlan    = null;
  protected static SelectFileWindowHandler  selectFileWindowHandler = null;


	/* Class Methods */
  protected static synchronized Plan getSelectFileWindowPlan()
  {
    if ( SelectFileWindowHandler.selectFileWindowPlan == null )
    {
      SelectFileWindowHandler.selectFileWindowPlan
	= WindowHandlerAbstractBaseClass.getPlan (
		SelectFileWindowHandler.SELECT_FILE_WINDOW_PLAN_NAME,
		"[SelectFileWindowHandler:getSelectFileWindowPlan]  "
		+ "Error:  Failure - Unable to access "
		+ "SELECT_FILE_WINDOW_PLAN_NAME" );
    }
    return SelectFileWindowHandler.selectFileWindowPlan;
  }

  public static SelectFileWindowHandler getSelectFileWindowHandler()
  {
    if ( SelectFileWindowHandler.selectFileWindowHandler == null )
    {
      SelectFileWindowHandler.selectFileWindowHandler
	= new SelectFileWindowHandler();
    }
    return SelectFileWindowHandler.selectFileWindowHandler;
  }




	/* Instance Variables */
  protected Button    newFileButton,
                      cancelButton,
		      selectButton;
  protected DataFile  selectedDataFile;


	/* Constructors */
  public SelectFileWindowHandler ( )
			{ createAndInitialize ( false ); }

  public SelectFileWindowHandler ( boolean  theUseExternalWindow )
			{ createAndInitialize ( theUseExternalWindow ); }

	/* Constructors can't be synchronized. */
	/* And this needs to be synchronized since we use the static plans */
  private synchronized
      boolean createAndInitialize ( boolean theUseExternalWindow )
  {
	/* Initialize selectedDataFile */
    selectedDataFile = null;

	/* Let the super-class do it's thing... */
    if ( super . createAndInitialize (
			     SelectFileWindowHandler.getSelectFileWindowPlan(),
			     SelectFileWindowHandler.SELECT_COMMAND,
			     null,
			     theUseExternalWindow,
			     false,
			     SelectFileWindowHandler.TITLE,
			     "SelectFileWindowHandler",
			     null )
	 == false )
    {
      return false;
    }

	/* Lets find our instance variables */
    newFileButton
      = (Button)
         SelectFileWindowHandler.getSelectFileWindowPlan().componentNamed (
			           SelectFileWindowHandler . NEW_FILE_BUTTON );
    cancelButton
      = (Button)
         SelectFileWindowHandler.getSelectFileWindowPlan().componentNamed (
				     SelectFileWindowHandler . CANCEL_BUTTON );
    selectButton
      = (Button)
         SelectFileWindowHandler.getSelectFileWindowPlan().componentNamed (
				     SelectFileWindowHandler . SELECT_BUTTON );

	/* Error-check */
    if (   ( newFileButton == null )
	|| ( cancelButton  == null )
	|| ( selectButton  == null ) )
    {
      System.err.println ( "[SelectFileWindowHandler:createAndInitialize]  "
	   + "Warning:  One or more Plan objects is NULL!!!\n"
	   + "    newFileButton...: " + newFileButton + "\n"
	   + "    cancelButton....: " + cancelButton  + "\n"
	   + "    selectButton....: " + selectButton  + "\n"
	   + "------------------------------------------------------------" );
    }


	/* Reset the plan so that we can use it again later on... */
    SelectFileWindowHandler.getSelectFileWindowPlan() . releaseObjects();

	/* Reconfigure our objects... */


	/* Report that we succeeded. */
    return true;
  }


  public void runModal ( )
  {
   	/* Re-Initialize selectedDataFile */
    selectedDataFile = null;

	/* And run modally */
    super . runModal();
  }


  public void run ( )
  {
   	/* Re-Initialize selectedDataFile */
    selectedDataFile = null;

	/* And run modally */
    super . run();
  }


  public Button      getNewFileButton()     { return newFileButton;    }
  public Button      getCancelButton()      { return cancelButton;     }
  public Button      getSelectButton()      { return selectButton;     }
  public DataFile    getSelectedDataFile()  { return selectedDataFile; }


	/* ExtendedTarget Interface */
  public boolean canPerformCommand ( String theCommand ) 
  {
    return 
      (    theCommand . equals ( SelectFileWindowHandler.CANCEL_COMMAND   )
	|| theCommand . equals ( SelectFileWindowHandler.SELECT_COMMAND   )
	|| theCommand . equals ( SelectFileWindowHandler.NEW_FILE_COMMAND )
	|| super . canPerformCommand ( theCommand ) );
  }


	/* Target Interface */
  public void performCommand ( String theCommand, Object theData )
  {
	/* Handle canceling... */
    if ( theCommand . equals( SelectFileWindowHandler.CANCEL_COMMAND ) )
    {
      selectedDataFile = null;   /* Null (Ie: NO) choice made. */
      getWindow() . hide();      /* Close the select-file window... */
    }

	/* Handle successful conclusion. */
    else if ( theCommand . equals ( SelectFileWindowHandler.SELECT_COMMAND ) )
    {
      if (   ( getFileListView() . selectedItem() != null )
	  && ( getFileListView() . count()         > 0    ) )
      {
	selectedDataFile
	  = (DataFile) (getFileListView() . selectedItem() . data());

	getWindow() . hide();  /* Close the select-file window... */
      }
      else
      {
	runAlert ( "Error:", "No File has been Selected.",
		   "OK", null, null );
      }
    }


	/* Handle a new-file choice */
    else if ( theCommand . equals( SelectFileWindowHandler.NEW_FILE_COMMAND ) )
    {
      FileChooser  fileChooser = new FileChooser ( getRootView(),
						   "Select File",
						   FileChooser.SAVE_TYPE );
      if ( WindowHandlerAbstractBaseClass.getPriorSelectedDirectory() != null )
      {
	fileChooser . setDirectory (
		  WindowHandlerAbstractBaseClass.getPriorSelectedDirectory() );
      }

      fileChooser . showModally();

      WindowHandlerAbstractBaseClass.setPriorSelectedDirectory (
						   fileChooser . directory() );

	/* Did the user "CANCEL" the fileChooser? */
      if ( fileChooser . file() == null )
	return;

      selectedDataFile = WindowHandlerAbstractBaseClass.validateNewFilename (
				fileChooser.directory() + fileChooser.file(),
				getRootView(), true );

	/* If user had a good file.... */
      if ( getSelectedDataFile() != null )
	getWindow() . hide();  /* Close the select-file window... */
    }

	/* Otherwise, let the superclass deal with it... */
    else
    {
      super . performCommand ( theCommand, theData );
    }

  } /* public void performCommand ( String theCommand, Object theData ) */

}
